﻿// Kerbal Engineer Redux
// Author:  CYBUTEK
// License: Attribution-NonCommercial-ShareAlike 3.0 Unported
//
// This class has taken a lot of inspiration from r4m0n's MuMech FuelFlowSimulator.  Although extremely
// similar to the code used within MechJeb, it is a clean re-write.  The similarities are a testiment
// to how well the MuMech code works and the robustness of the simulation algorithem used.

using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Engineer
{
    class Simulator
    {
        List<PartSim> partSims;
        int currentStage = 0;

        public Stage[] RunSimulation(List<Part> parts, float gravity, float atmosphere)
        {
            BuildVessel(parts, atmosphere);

            currentStage = Staging.lastStage;
            Stage[] stages = new Stage[currentStage + 1];

            while (currentStage >= 0)
            {
                Stage stage = new Stage();
                double stageTime = 0;
                double stageDeltaV = 0;
                double totalStageThrust = 0;
                double totalStageActualThrust = 0;
                double totalStageIsp = 0;
                foreach (PartSim engine in ActiveEngines)
                {
                    totalStageActualThrust += engine.actualThrust;
                    totalStageThrust += engine.thrust;
                    totalStageIsp += engine.isp * engine.thrust;
                }

                stage.thrust = totalStageThrust;
                stage.thrustToWeight = (float)(totalStageThrust / (ShipMass * gravity));
                stage.actualThrust = totalStageActualThrust;
                stage.actualThrustToWeight = (float)(totalStageActualThrust / (ShipMass * gravity));

                if (totalStageThrust > 0 && totalStageIsp > 0)
                {
                    stage.isp = totalStageIsp / totalStageThrust;
                }

                foreach (PartSim partSim in partSims)
                {
                    if (partSim.decoupledInStage == currentStage - 1)
                    {
                        stage.cost += partSim.part.partInfo.cost;
                        stage.mass += partSim.GetMass(currentStage);
                    }
                }

                while (!AllowedToStage())
                {
                    totalStageThrust = 0;
                    foreach (PartSim engine in ActiveEngines)
                    {
                        if (engine.actualThrust > 0)
                        {
                            totalStageThrust += engine.actualThrust;
                        }
                        else
                        {
                            totalStageThrust += engine.thrust;
                        }
                        
                    }

                    SetResourceDrainRates();

                    double resourceDrainTime = float.MaxValue;
                    foreach (PartSim partSim in partSims)
                    {
                        double time = partSim.TimeToDrainResource();
                        if (time < resourceDrainTime)
                        {
                            resourceDrainTime = time;
                        }
                    }

                    double startMass = ShipMass;
                    foreach (PartSim partSim in partSims)
                    {
                        partSim.DrainResources(resourceDrainTime);
                    }
                    double endMass = ShipMass;
                    stageTime += resourceDrainTime;

                    if (resourceDrainTime > 0 && startMass > endMass && startMass > 0 && endMass > 0)
                    {
                        stageDeltaV += totalStageThrust * resourceDrainTime / (startMass - endMass) * Math.Log(startMass / endMass);
                    }
                }

                stage.deltaV = stageDeltaV;
                stage.time = stageTime;
                stages[currentStage] = stage;

                currentStage--;
                ActivateStage();
            }

            for (int i = 0; i < stages.Length; i++)
            {
                for (int j = i; j >= 0; j--)
                {
                    stages[i].totalCost += stages[j].cost;
                    stages[i].totalMass += stages[j].mass;
                    stages[i].totalDeltaV += stages[j].deltaV;
                    stages[i].totalTime += stages[j].time;
                }
                for (int j = i; j < stages.Length; j++)
                {
                    stages[i].inverseTotalDeltaV += stages[j].deltaV;
                }
            }

            return stages;
        }

        private void BuildVessel(List<Part> parts, float atmosphere)
        {
            partSims = new List<PartSim>();
            Dictionary<Part, PartSim> partSimLookup = new Dictionary<Part, PartSim>();

            foreach (Part part in parts)
            {
                PartSim partSim = new PartSim(part, atmosphere);
                partSims.Add(partSim);
                partSimLookup.Add(part, partSim);
            }

            foreach (PartSim partSim in partSims)
            {
                partSim.SetSourceNodes(partSimLookup);
            }
        }

        private bool AllowedToStage()
        {
            List<PartSim> engines = ActiveEngines;

            if (engines.Count == 0)
            {
                return true;
            }

            foreach (PartSim partSim in partSims)
            {
                if (partSim.decoupledInStage == (currentStage - 1))
                {
                    if (!partSim.Resources.Empty || (engines.Contains(partSim) && !partSim.IsSepratron))
                    {
                        return false;
                    }
                }
            }

            if (currentStage > 0)
            {
                return true;
            }

            return false;
        }

        private void SetResourceDrainRates()
        {
            foreach (PartSim partSim in partSims)
            {
                partSim.ResourceDrains.Reset();
            }

            List<PartSim> engines = ActiveEngines;

            foreach (PartSim engine in engines)
            {
                engine.SetResourceDrainRates(partSims);
            }
        }

        private void ActivateStage()
        {
            List<PartSim> decoupledParts = new List<PartSim>();

            foreach (PartSim partSim in partSims)
            {
                if (partSim.decoupledInStage == currentStage)
                {
                    decoupledParts.Add(partSim);
                }
            }

            foreach (PartSim partSim in decoupledParts)
            {
                partSims.Remove(partSim);
            }

            foreach (PartSim partSim in partSims)
            {
                foreach (PartSim decoupledPart in decoupledParts)
                {
                    partSim.RemoveSourcePart(decoupledPart);
                }
            }
        }

        private List<PartSim> ActiveEngines
        {
            get
            {
                List<PartSim> engines = new List<PartSim>();
                {
                    foreach (PartSim partSim in partSims)
                    {
                        if (partSim.IsEngine && partSim.InverseStage >= currentStage && partSim.CanDrawNeededResources(partSims))
                        {
                            engines.Add(partSim);
                        }
                    }
                }

                return engines;
            }
        }

        private double ShipMass
        {
            get
            {
                double mass = 0f;

                foreach (PartSim partSim in partSims)
                {
                    mass += partSim.GetMass(currentStage);
                }

                return mass;
            }
        }
    }
}
